"""Library exceptions."""


class MissingRequiredArgumentError(Exception):
    """There is a missing argument in query."""


class UnusedArgumentsError(Exception):
    """There are unused arguments in query."""



"""Module with extensions of asyncpg `Connection` class."""
import functools
import typing

import asyncpg
from asyncpg import cursor

from asyncpgx import prepared_statement
from asyncpgx import query as query_module


class ConnectionX(asyncpg.connection.Connection):
    """Extended version of asyncpg `Connection` class.

    Provides various extension methods, but doesn't touches the original
    ones
    """

    def _prepare_asyncpg_parameters(
        self, query: str, args: typing.Any, converter: query_module.QueryParamsConverter
    ) -> typing.Tuple[str, typing.List]:
        """Prepare high-level query and arguments to underlying asyncpg
        backend."""
        converted_query, params_order_list = query_module.construct_asyncpg_query(query)
        return converted_query, converter.prepare_asyncpg_args(args, params_order_list)

    async def named_execute(self, query: str, args: typing.Dict, timeout: typing.Optional[float] = None) -> str:
        """Extended versions of `execute` with support of the named parameters.

        :param query: SQL query to execute (could include named parameters).
        :param args: Dict with the parameters values.
        :param timeout: Optional timeout value in seconds.
        """
        converted_query, asyncpg_args = self._prepare_asyncpg_parameters(
            query, args, query_module.QueryParamsDictConverter()
        )
        query_result: str = await super().execute(converted_query, *asyncpg_args, timeout=timeout)
        return query_result

    async def named_executemany(self, query: str, args: typing.List, *, timeout: typing.Optional[float] = None) -> None:
        """Extended versions of `executemany` with support of the named
        parameters.

        :param query: SQL query to execute (could include named parameters).
        :param args: List of dicts with the parameters values.
        :param timeout: Optional timeout value in seconds.
        """
        converted_query, asyncpg_args = self._prepare_asyncpg_parameters(
            query, args, query_module.QueryParamsListDictConverter()
        )
        query_result: None = await super().executemany(converted_query, asyncpg_args, timeout=timeout)
        return query_result

    async def named_fetch(
        self, query: str, args: typing.Dict, timeout: typing.Optional[float] = None
    ) -> typing.List[asyncpg.Record]:
        """Extended versions of `fetch` with support of the named parameters.

        :param query: SQL query to execute (could include named parameters).
        :param args: Dict with the parameters values.
        :param timeout: Optional timeout value in seconds.
        """
        converted_query, asyncpg_args = self._prepare_asyncpg_parameters(
            query, args, query_module.QueryParamsDictConverter()
        )
        query_result: typing.List[asyncpg.Record] = await super().fetch(converted_query, *asyncpg_args, timeout=timeout)
        return query_result

    async def named_fetchval(
        self, query: str, args: typing.Dict, column: int = 0, timeout: typing.Optional[float] = None
    ) -> typing.Optional[typing.Any]:
        """Extended versions of `fetchval` with support of the named
        parameters.

        :param query: SQL query to execute (could include named parameters).
        :param args: Dict with the parameters values.
        :param column: Numeric index within the record of the value to return.
        :param timeout: Optional timeout value in seconds.
        """
        converted_query, asyncpg_args = self._prepare_asyncpg_parameters(
            query, args, query_module.QueryParamsDictConverter()
        )
        return await super().fetchval(converted_query, *asyncpg_args, column=column, timeout=timeout)

    async def named_fetchrow(
        self, query: str, args: typing.Dict, timeout: typing.Optional[float] = None
    ) -> typing.Optional[asyncpg.Record]:
        """Extended versions of `fetchrow` with support of the named
        parameters.

        :param query: SQL query to execute (could include named parameters).
        :param args: Dict with the parameters values.
        :param timeout: Optional timeout value in seconds.
        """
        converted_query, asyncpg_args = self._prepare_asyncpg_parameters(
            query, args, query_module.QueryParamsDictConverter()
        )
        return await super().fetchrow(converted_query, *asyncpg_args, timeout=timeout)

    def named_cursor(
        self,
        query: str,
        args: typing.Dict,
        prefetch: typing.Optional[int] = None,
        timeout: typing.Optional[float] = None,
    ) -> cursor.CursorFactory:
        """Extended version of `cursor` with support of the named parameters.

        :param query: SQL query to execute (could include named parameters).
        :param args: Dict with the parameters values.
        :param prefetch: The number of rows the *cursor iterator* will prefetch (defaults to ``50``.)
        :param timeout: Optional timeout value in seconds.
        """
        converted_query, asyncpg_args = self._prepare_asyncpg_parameters(
            query, args, query_module.QueryParamsDictConverter()
        )
        return super().cursor(converted_query, *asyncpg_args, prefetch=prefetch, timeout=timeout)

    async def named_prepare(
        self, query: str, *, timeout: typing.Optional[float] = None
    ) -> prepared_statement.PreparedStatementX:
        """Extended version of `prepare` with support of the named parameters.

        :param query: SQL query to execute (could include named parameters).
        :param timeout: Optional timeout value in seconds.
        """
        converted_query, params_order_list = query_module.construct_asyncpg_query(query)
        self._check_open()
        stmt = await self._get_statement(converted_query, timeout, named=True, use_cache=False)
        return prepared_statement.PreparedStatementX(self, converted_query, stmt, query, params_order_list)


create_pool = functools.partial(asyncpg.create_pool, connection_class=ConnectionX)
connect = functools.partial(asyncpg.connect, connection_class=ConnectionX)



"""Module with tools for queries processing."""
import abc
import re
import typing

from asyncpgx import exceptions


PARAMS_REGEXP = re.compile(r"(?<![:\w\x5c]):(\w+)(?!:)", re.UNICODE)


def construct_asyncpg_query(query: str) -> typing.Tuple[str, typing.List]:
    """Construct asyncpg query from high-level one."""
    i = 1
    params_order_list = []

    def _construct_replacement(match_obj: typing.Match) -> str:
        nonlocal i
        new_numeric_param = f'${i}'
        params_order_list.append(match_obj.group(0)[1::])
        i += 1
        return new_numeric_param

    query = PARAMS_REGEXP.sub(_construct_replacement, query)
    return query, params_order_list


# pylint: disable=too-few-public-methods
class QueryParamsConverter(abc.ABC):
    """Abstract class for converting our high-level API to low-level asyncpg
    API."""

    @abc.abstractmethod
    def prepare_asyncpg_args(self, original_args: typing.Any, params_order_list: typing.List) -> typing.List:
        """Prepare asyncpg method arguments."""
        raise NotImplementedError()  # pragma: no cover


class QueryParamsListDictConverter(QueryParamsConverter):
    """Converts list of dicts named parameters to the asyncpg-friendly
    format."""

    def prepare_asyncpg_args(
        self, original_args: typing.List[typing.Dict], params_order_list: typing.List
    ) -> typing.List:
        """Prepare asyncpg method arguments."""
        asyncpg_args = []
        for arg in original_args:
            asyncpg_args.append(QueryParamsDictConverter().prepare_asyncpg_args(arg, params_order_list))

        return asyncpg_args


class QueryParamsDictConverter(QueryParamsConverter):
    """Converts dict named parameters to the asyncpg-friendly format."""

    def prepare_asyncpg_args(self, original_args: typing.Dict, params_order_list: typing.List) -> typing.List:
        """Prepare asyncpg method arguments."""
        asyncpg_args = []
        used_arguments = set()
        for param in params_order_list:
            try:
                asyncpg_args.append(original_args[param])
            except KeyError as exc:
                raise exceptions.MissingRequiredArgumentError(f'Missing required argument: {param}') from exc

            used_arguments.add(param)

        if len(used_arguments) != len(original_args):
            unused_arguments = set(original_args.keys()).difference(used_arguments)
            raise exceptions.UnusedArgumentsError(f'Arguments: {unused_arguments} are unused')

        return asyncpg_args



"""Module with extensions of asyncpg `PreparedStatement` class."""
import typing

import asyncpg
import asyncpg.cursor
import asyncpg.prepared_stmt

from asyncpgx import query as query_module


class PreparedStatementX(asyncpg.prepared_stmt.PreparedStatement):
    """Extended version of asyncpg `PreparedStatement` class.

    Provides various extension methods, but doesn't touches the original
    ones
    """

    __slots__ = ('_original_query', '_params_order_list')

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        connection: asyncpg.Connection,
        query: str,
        state: typing.Any,  # i couldn't import PreparedStatementState
        original_query: str,
        params_order_list: typing.List,
    ):
        super().__init__(connection, query, state)
        self._original_query = original_query
        self._params_order_list = params_order_list

    def named_cursor(
        self,
        args: typing.Dict,
        prefetch: typing.Optional[int] = None,
        timeout: typing.Optional[float] = None,
    ) -> asyncpg.cursor.CursorFactory:
        """Extended version of `cursor` with support of the named parameters.

        :param args: Dict with the parameters values.
        :param prefetch: The number of rows the *cursor iterator* will prefetch (defaults to ``50``.)
        :param timeout: Optional timeout value in seconds.
        """
        prepared_args = query_module.QueryParamsDictConverter().prepare_asyncpg_args(args, self._params_order_list)
        return super().cursor(*prepared_args, prefetch=prefetch, timeout=timeout)

    async def named_fetch(
        self, args: typing.Dict, timeout: typing.Optional[float] = None
    ) -> typing.List[asyncpg.Record]:
        """Extended version of `fetch` with support of the named parameters.

        :param args: Dict with the parameters values.
        :param timeout: Optional timeout value in seconds.
        """
        prepared_args = query_module.QueryParamsDictConverter().prepare_asyncpg_args(args, self._params_order_list)
        query_result: typing.List[asyncpg.Record] = await super().fetch(*prepared_args, timeout=timeout)
        return query_result

    async def named_fetchval(
        self, args: typing.Dict, column: int = 0, timeout: typing.Optional[float] = None
    ) -> typing.Any:
        """Extended version of `fetchval` with support of the named parameters.

        :param args: Dict with the parameters values.
        :param column: Numeric index within the record of the value to return.
        :param timeout: Optional timeout value in seconds.
        """
        prepared_args = query_module.QueryParamsDictConverter().prepare_asyncpg_args(args, self._params_order_list)
        return await super().fetchval(*prepared_args, column=column, timeout=timeout)

    async def named_fetchrow(self, args: typing.Dict, timeout: typing.Optional[float] = None) -> asyncpg.Record:
        """Extended version of `fetchrow` with support of the named parameters.

        :param args: Dict with the parameters values.
        :param timeout: Optional timeout value in seconds.
        """
        prepared_args = query_module.QueryParamsDictConverter().prepare_asyncpg_args(args, self._params_order_list)
        return await super().fetchrow(*prepared_args, timeout=timeout)



"""Test `connection` module."""
import pytest

from asyncpgx import connection as connection_module
from asyncpgx import exceptions


@pytest.mark.asyncio
async def test_named_execute_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_execute` method with named parameters."""
    await postgres_connection.named_execute(
        'INSERT INTO test(id, test_1, test_2) VALUES (:id, :test_1, :test_2);', {'id': 1, 'test_1': '1', 'test_2': '2'}
    )
    fetch_result = await postgres_connection.fetchrow('SELECT id, test_1, test_2 FROM test WHERE id=1;')

    assert fetch_result['id'] == 1
    assert fetch_result['test_1'] == '1'
    assert fetch_result['test_2'] == '2'


@pytest.mark.asyncio
async def test_named_executemany_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_executemany` method with named parameters."""
    await postgres_connection.named_executemany(
        'INSERT INTO test(id, test_1, test_2) VALUES (:id, :test_1, :test_2);',
        [
            {'id': 1, 'test_1': '1', 'test_2': '1'},
            {'id': 2, 'test_1': '2', 'test_2': '2'},
            {'id': 3, 'test_1': '3', 'test_2': '3'},
        ],
    )
    fetch_result = await postgres_connection.fetch('SELECT id, test_1, test_2 FROM test ORDER BY id ASC;')

    for i, row in enumerate(fetch_result):
        assert row['id'] == i + 1
        assert row['test_1'] == str(i + 1)
        assert row['test_2'] == str(i + 1)


@pytest.mark.asyncio
async def test_named_fetch_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_fetch` method with named parameters."""
    await postgres_connection.execute(
        '''
        INSERT INTO test(id, test_1, test_2) VALUES (1, '1', '1'),(2, '2', '2'),(3, '3', '3')
        '''
    )

    fetch_result = await postgres_connection.named_fetch('SELECT id, test_1, test_2 FROM test WHERE id=:id;', {'id': 2})

    assert fetch_result[0]['id'] == 2
    assert fetch_result[0]['test_1'] == '2'
    assert fetch_result[0]['test_2'] == '2'


@pytest.mark.asyncio
async def test_named_fetchval_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_fetchval` method with named parameters."""
    await postgres_connection.execute(
        '''
        INSERT INTO test(id, test_1, test_2) VALUES (1, '1', '1'),(2, '2', '2'),(3, '3', '3')
        '''
    )

    fetch_result = await postgres_connection.named_fetchval(
        'SELECT id, test_1, test_2 FROM test WHERE id=:id;', {'id': 2}, column=0
    )

    assert fetch_result == 2


@pytest.mark.asyncio
async def test_named_fetchrow_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_fetchrow` method with named parameters."""
    await postgres_connection.execute(
        '''
        INSERT INTO test(id, test_1, test_2) VALUES (1, '1', '1'),(2, '2', '2'),(3, '3', '3')
        '''
    )

    fetch_result = await postgres_connection.named_fetchrow(
        'SELECT id, test_1, test_2 FROM test WHERE id=:id;',
        {'id': 2},
    )

    assert fetch_result and fetch_result['id'] == 2


@pytest.mark.asyncio
async def test_named_cursor_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_cursor` method with named parameters."""
    await postgres_connection.execute(
        '''
        INSERT INTO test(id, test_1, test_2) VALUES (1, '1', '1'),(2, '2', '2'),(3, '2', '3')
        '''
    )

    async with postgres_connection.transaction():
        async for row in postgres_connection.named_cursor(
            'SELECT id, test_1, test_2 FROM test WHERE test_1=:test_1', {'test_1': '2'}
        ):
            assert row['test_1'] == '2'
            assert row['id'] in {2, 3}


@pytest.mark.asyncio
async def test_named_prepare_cursor_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_prepare` with `named_cursor` method with named
    parameters."""
    await postgres_connection.execute(
        '''INSERT INTO test (id, test_1, test_2) VALUES (1, '1', '3'),
                                                        (2, '2', '4'),
                                                        (3, '2', '5')'''
    )

    prepared_statement = await postgres_connection.named_prepare(
        '''SELECT id, test_1, test_2 FROM test WHERE test_1=:test_1;'''
    )

    async with postgres_connection.transaction():
        async for row in prepared_statement.named_cursor({'test_1': '2'}):
            assert row['test_1'] == '2'
            assert row['id'] in {2, 3}


@pytest.mark.asyncio
async def test_named_prepare_fetch_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_prepare` with `named_fetch` method with named parameters."""
    await postgres_connection.execute('''INSERT INTO test (id, test_1, test_2) VALUES (1, '2', '3')''')
    prepared_statement = await postgres_connection.named_prepare(
        '''SELECT id, test_1, test_2 FROM test WHERE id=:id;'''
    )

    query_result = await prepared_statement.named_fetch({'id': 1})

    assert query_result[0]['id'] == 1
    assert query_result[0]['test_1'] == '2'
    assert query_result[0]['test_2'] == '3'


@pytest.mark.asyncio
async def test_named_prepare_fetchval_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_prepare` with `named_fetchval` method with named
    parameters."""
    await postgres_connection.execute('''INSERT INTO test (id, test_1, test_2) VALUES (1, '2', '3'), (2, '4', '5')''')
    prepared_statement = await postgres_connection.named_prepare(
        '''SELECT id, test_1, test_2 FROM test WHERE id=:id;'''
    )

    query_result = await prepared_statement.named_fetchval({'id': 2}, column=1)

    assert query_result == '4'


@pytest.mark.asyncio
async def test_named_prepare_fetchrow_with_named_parameters(postgres_connection: connection_module.ConnectionX) -> None:
    """Test `named_prepare` with `named_fetchrow` method with named
    parameters."""
    await postgres_connection.execute('''INSERT INTO test (id, test_1, test_2) VALUES (1, '2', '3'), (2, '4', '5')''')
    prepared_statement = await postgres_connection.named_prepare(
        '''SELECT id, test_1, test_2 FROM test WHERE id=:id;'''
    )

    query_result = await prepared_statement.named_fetchrow({'id': 2})

    assert query_result['id'] == 2
    assert query_result['test_1'] == '4'
    assert query_result['test_2'] == '5'


@pytest.mark.asyncio
async def test_missing_arguments(postgres_connection: connection_module.ConnectionX) -> None:
    """Test missing arguments exception."""
    with pytest.raises(exceptions.MissingRequiredArgumentError):
        await postgres_connection.named_execute(
            '''SELECT id, test_1, test_2 FROM test WHERE id=:id AND test_1=:test_1;''', {'id': 1}
        )


@pytest.mark.asyncio
async def test_unused_arguments(postgres_connection: connection_module.ConnectionX) -> None:
    """Test unused arguments exception."""
    with pytest.raises(exceptions.UnusedArgumentsError):
        await postgres_connection.named_execute(
            '''SELECT id, test_1, test_2 FROM test WHERE id=:id;''', {'id': 1, 'test_1': '2'}
        )


@pytest.mark.asyncio
async def test_missing_arguments_list(postgres_connection: connection_module.ConnectionX) -> None:
    """Test missing arguments exception with list of params."""
    with pytest.raises(exceptions.MissingRequiredArgumentError):
        await postgres_connection.named_executemany(
            '''SELECT id, test_1, test_2 FROM test WHERE id=:id AND test_1=:test_1;''', [{'id': 1}, {'id': 2}]
        )


@pytest.mark.asyncio
async def test_unused_arguments_list(postgres_connection: connection_module.ConnectionX) -> None:
    """Test unused arguments exception with list of params."""
    with pytest.raises(exceptions.UnusedArgumentsError):
        await postgres_connection.named_executemany(
            '''SELECT id, test_1, test_2 FROM test WHERE id=:id;''',
            [{'id': 1, 'test_1': '2'}, {'id': 2, 'test_1': '3'}],
        )



"""Tests for `query` module."""
from asyncpgx import query


def test_construct_query():
    """Test converter construct query."""
    new_query, params_order_list = query.construct_asyncpg_query(
        '''SELECT * FROM some_table WHERE id=:id AND some_field_1=:some_field_1 AND some_field_2=:some_field_2;'''
    )

    expected_new_query = '''SELECT * FROM some_table WHERE id=$1 AND some_field_1=$2 AND some_field_2=$3;'''
    assert new_query == expected_new_query
    assert params_order_list == ['id', 'some_field_1', 'some_field_2']


def test_list_dict_converter_prepare_asyncpg_args():
    """Test `prepare_asyncpg_args` for list dict converter."""
    converter = query.QueryParamsListDictConverter()

    asyncpg_args = converter.prepare_asyncpg_args(
        [
            {'some_field_1': '1', 'some_field_2': '2', 'some_field_3': '3'},
            {'some_field_1': '4', 'some_field_2': '5', 'some_field_3': '6'},
            {'some_field_1': '7', 'some_field_2': '8', 'some_field_3': '9'},
        ],
        ['some_field_1', 'some_field_2', 'some_field_3'],
    )

    assert asyncpg_args == [['1', '2', '3'], ['4', '5', '6'], ['7', '8', '9']]


def test_dict_converter_prepare_asyncpg_args():
    """Test `prepare_asyncpg_args` for dict converter."""
    converter = query.QueryParamsDictConverter()

    asyncpg_args = converter.prepare_asyncpg_args(
        {'some_field_1': '1', 'some_field_2': '2', 'some_field_3': '3'},
        ['some_field_1', 'some_field_2', 'some_field_3'],
    )

    assert asyncpg_args == ['1', '2', '3']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asyncpgx', 'asyncpgx.tests']

package_data = \
{'': ['*']}

install_requires = \
['asyncpg>=0.21,<0.27']

setup_kwargs = {
    'name': 'asyncpgx',
    'version': '1.3.0',
    'description': 'User-friendly extensions for asyncpg',
    'long_description': "# asyncpgx\n[![Build passed](https://img.shields.io/github/workflow/status/laukhin/asyncpgx/CI)](https://github.com/laukhin/asyncpgx/actions?query=workflow%3ACI)\n[![Test coverage](https://img.shields.io/codecov/c/github/laukhin/asyncpgx)](https://codecov.io/gh/laukhin/asyncpgx)\n[![Version](https://img.shields.io/pypi/v/asyncpgx)](https://pypi.org/project/asyncpgx/)\n\nExtensions for asyncpg.\n\nBased on the [asyncpg](https://github.com/MagicStack/asyncpg) and highly inspired by the [sqlx](https://github.com/jmoiron/sqlx) package\n\nThis package supports 3.6+ python versions\n\n## Setup\nUse `pip install asyncpgx`\n\n## Purpose\nThis is a thin wrapper on the `asyncpg` package.\nOur purpose is to provide convenient extensions to the original package.\nWe're trying to delegate as much work as we can to the asyncpg (basically our extension methods are high-level proxies to the underlying ones)\nand make only converting job.\nOriginal asyncpg API stays the same, you can see it in the [asyncpg documentation](https://magicstack.github.io/asyncpg/current/).\n\n## Functionality\n* queries with named parameters, i.e.\n```python\nimport asyncpgx\n\nconnection = await asyncpgx.connect('postgresql://127.0.0.1:5432')\nawait connection.named_fetch('''SELECT field FROM some_table WHERE id <= :id;''', {'id': 1})\n```\n* prepared statements with named parameters, i.e.\n```python\nimport asyncpgx\n\nconnection = await asyncpgx.connect('postgresql://127.0.0.1:5432')\nprepared_statement = await connection.named_prepare('''SELECT field FROM some_table WHERE id <= :id;''')\nawait prepared_statement.named_fetch({'id': 1})\n```\n\n## Documentation\nYou can find project documentation [here](https://laukhin.github.io/asyncpgx/index.html)\n\n## Changelog\nYou can find all releases description [here](https://github.com/laukhin/asyncpgx/releases)\n",
    'author': 'Vladislav Laukhin',
    'author_email': 'laukhin97@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/laukhin/asyncpgx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<=3.10',
}


setup(**setup_kwargs)



"""Common fixtures."""
import os

import pytest

from asyncpgx import connection as connection_module


POSTGRES_DSN = os.getenv('POSTGRES_TEST_DSN', 'postgresql://127.0.0.1:5432')


@pytest.fixture
async def postgres_connection():
    """Fixture which establishes the connection, creates the test table and
    cleans everything on scope close."""
    connection = await connection_module.connect(POSTGRES_DSN)
    await connection.execute('''CREATE TABLE test (id int PRIMARY KEY, test_1 varchar (256), test_2 varchar (256))''')
    yield connection
    await connection.execute('''DROP TABLE test CASCADE;''')
